""" Definitions for Omada enums. """

from enum import IntEnum


class DeviceStatus(IntEnum):
    """Known status codes for devices."""

    DISCONNECTED = 0
    DISCONNECTED_MIGRATING = 1
    PROVISIONING = 10
    CONFIGURING = 11
    UPGRADING = 12
    REBOOTING = 13
    CONNECTED = 14
    CONNECTED_WIRELESS = 15
    CONNECTED_MIGRATING = 16
    CONNECTED_WIRELESS_MIGRATING = 17
    PENDING = 20
    PENDING_WIRELESS = 21
    ADOPTING = 22
    ADOPTING_WIRELESS = 23
    ADOPT_FAILED = 24
    ADOPT_FAILED_WIRELESS = 25
    MANAGED_EXTERNALLY = 26
    MANAGED_EXTERNALLY_WIRELESS = 27
    HEARTBEAT_MISSED = 30
    HEARTBEAT_MISSED_WIRELESS = 31
    HEARTBEAT_MISSED_MIGRATING = 32
    HEARTBEAT_MISSED_WIRELESS_MIGRATING = 33
    ISOLATED = 40
    ISOLATED_MIGRATING = 41


class DeviceStatusCategory(IntEnum):
    """Known status categories for devices"""

    DISCONNECTED = 0
    CONNECTED = 1
    PENDING = 2
    HEARTBEAT_MISSED = 3
    ISOLATED = 4


class PortType(IntEnum):
    """Known types of switch port."""

    COPPER = 1
    COMBO = 2
    SFP = 3


class LinkStatus(IntEnum):
    """Known link statuses."""

    LINK_UP = 0
    LINK_DOWN = 2


class LinkSpeed(IntEnum):
    """Known link speeds."""

    SPEED_AUTO = 0
    SPEED_10_MBPS = 1
    SPEED_100_MBPS = 2
    SPEED_1_GBPS = 3
    SPEED_10_GBPS = 4


class LinkDuplex(IntEnum):
    """Known link duplex modes"""

    AUTO = 0
    HALF = 1
    FULL = 2


class Eth802Dot1X(IntEnum):
    """802.1x auth modes."""

    FORCE_UNAUTHORIZED = 0
    FORCE_AUTHORIZED = 1
    AUTO = 2


class BandwidthControl(IntEnum):
    """Modes of bandwidth control."""

    OFF = 0
    RATE_LIMIT = 1
    STORM_CONTROL = 2


class PoEMode(IntEnum):
    """Settings for PoE policy."""

    DISABLED = 0
    ENABLED = 1
    USE_DEVICE_SETTINGS = 2



"""Client for Omada Site requests."""

from typing import List, Optional, Union
from .omadaapiconnection import OmadaApiConnection

from .devices import (
    OmadaAccessPoint,
    OmadaDevice,
    OmadaFirmwareUpdate,
    OmadaListDevice,
    OmadaPortProfile,
    OmadaSwitch,
    OmadaSwitchPort,
    OmadaSwitchPortDetails,
    OmadaAccesPointLanPortSettings,
)

from .exceptions import (
    InvalidDevice,
)

from .definitions import BandwidthControl, Eth802Dot1X, LinkDuplex, LinkSpeed, PoEMode


class SwitchPortOverrides:
    """
    Overrides that can be applied to a switch port.

    Currently, we don't support bandwidth limits and mirroring modes.
    Due to the way the API works, we have to specify overrides for everything,
    we can't just override a single profile setting. Therefore, you may need to
    initialise all of these parameters to avoid overwriting settings.
    """

    def __init__(
        self,
        enable_poe: bool = True,
        dot1x_mode: Eth802Dot1X = Eth802Dot1X.FORCE_AUTHORIZED,
        duplex: LinkDuplex = LinkDuplex.AUTO,
        link_speed: LinkSpeed = LinkSpeed.SPEED_AUTO,
        lldp_med_enable: bool = True,
        loopback_detect: bool = True,
        spanning_tree_enable: bool = False,
        port_isolation: bool = False,
    ):
        self.enable_poe = enable_poe
        self.dot1x_mode = dot1x_mode
        self.duplex = duplex
        self.link_speed = link_speed
        self.lldp_med_enable = lldp_med_enable
        self.loopback_detect = loopback_detect
        self.spanning_tree_enable = spanning_tree_enable
        self.port_isolation = port_isolation


class AccessPointPortSettings:
    """
    Settings that can be applied to network ports on access points

    Specify the values you want to modify. The remaining values will be unaffected
    """

    def __init__(
        self,
        enable_poe: Optional[bool] = None,
        vlan_enable: Optional[bool] = None,
        vlan_id: Optional[int] = None,
    ):
        self.enable_poe = enable_poe
        self.vlan_enable = vlan_enable
        self.vlan_id = vlan_id


class OmadaSiteClient:
    """Client for querying an Omada site's devices."""

    def __init__(self, site_id: str, api: OmadaApiConnection):
        self._api = api
        self._site_id = site_id

    async def get_devices(self) -> List[OmadaListDevice]:
        """Get the list of devices on the site."""

        result = await self._api.request(
            "get", self._api.format_url("devices", self._site_id)
        )

        return [OmadaListDevice(d) for d in result]

    async def get_switches(self) -> List[OmadaSwitch]:
        """Get the list of switches on the site."""

        return [
            await self.get_switch(d)
            for d in await self.get_devices()
            if d.type == "switch"
        ]

    async def get_access_points(self) -> List[OmadaAccessPoint]:
        """Get the list of access points on the site."""

        return [
            await self.get_access_point(d)
            for d in await self.get_devices()
            if d.type == "ap"
        ]

    async def get_access_point(
        self, mac_or_device: Union[str, OmadaDevice]
    ) -> OmadaAccessPoint:
        """Get an access point by Mac address or Omada device."""

        if isinstance(mac_or_device, OmadaDevice):
            if mac_or_device.type != "ap":
                raise InvalidDevice()
            mac = mac_or_device.mac
        else:
            mac = mac_or_device

        result = await self._api.request(
            "get", self._api.format_url(f"eaps/{mac}", self._site_id)
        )

        return OmadaAccessPoint(result)

    async def get_switch(self, mac_or_device: Union[str, OmadaDevice]) -> OmadaSwitch:
        """Get a switch by Mac address or Omada device."""

        if isinstance(mac_or_device, OmadaDevice):
            if mac_or_device.type != "switch":
                raise InvalidDevice()
            mac = mac_or_device.mac
        else:
            mac = mac_or_device

        result = await self._api.request(
            "get", self._api.format_url(f"switches/{mac}", self._site_id)
        )

        return OmadaSwitch(result)

    async def get_switch_ports(
        self, mac_or_device: Union[str, OmadaDevice]
    ) -> List[OmadaSwitchPortDetails]:
        """Get a switch by Mac address or Omada device."""

        if isinstance(mac_or_device, OmadaDevice):
            if mac_or_device.type != "switch":
                raise InvalidDevice()
            mac = mac_or_device.mac
        else:
            mac = mac_or_device

        result = await self._api.request(
            "get", self._api.format_url(f"switches/{mac}/ports", self._site_id)
        )

        return [OmadaSwitchPortDetails(p) for p in result]

    async def get_switch_port(
        self,
        mac_or_device: Union[str, OmadaDevice],
        index_or_port: Union[int, OmadaSwitchPort],
    ) -> OmadaSwitchPortDetails:
        """Get a switch by Mac address or Omada device."""

        if isinstance(mac_or_device, OmadaDevice):
            if mac_or_device.type != "switch":
                raise InvalidDevice()
            mac = mac_or_device.mac
        else:
            mac = mac_or_device

        if isinstance(index_or_port, OmadaSwitchPort):
            port = index_or_port.port
        else:
            port = index_or_port

        result = await self._api.request(
            "get", self._api.format_url(f"switches/{mac}/ports/{port}", self._site_id)
        )

        return OmadaSwitchPortDetails(result)

    async def update_access_point_port(
        self,
        mac_or_device: Union[str, OmadaDevice],
        port_name: str,
        setting: AccessPointPortSettings,
    ) -> OmadaAccesPointLanPortSettings:
        """Update the settings for a lan port on the access point."""

        # Get the latest representation of the acccess point
        access_point = await self.get_access_point(mac_or_device)

        port_settings = [
            {
                "id": port_name,
                "lanPort": port_name,
                "localVlanEnable": setting.vlan_enable
                if setting.vlan_enable is not None
                else ps.local_vlan_enable,
                "localVlanId": setting.vlan_id
                if setting.vlan_id is not None
                else ps.local_vlan_id,
                "poeOutEnable": setting.enable_poe
                if setting.enable_poe is not None and ps.supports_poe
                else ps.poe_enable,
            }
            for ps in access_point.lan_port_settings
            if ps.port_name == port_name
        ]

        payload = {"lanPortSettings": port_settings}

        result = await self._api.request(
            "patch",
            self._api.format_url(f"eaps/{access_point.mac}", self._site_id),
            payload=payload,
        )

        updated_ap = OmadaAccessPoint(result)
        # The caller probably only cares about the updated port status
        return next(p for p in updated_ap.lan_port_settings if p.port_name == port_name)

    async def update_switch_port(
        self,
        mac_or_device: Union[str, OmadaDevice],
        index_or_port: Union[int, OmadaSwitchPort],
        new_name: Optional[str] = None,
        profile_id: Optional[str] = None,
        overrides: Optional[SwitchPortOverrides] = None,
    ) -> OmadaSwitchPortDetails:
        """Applies an existing profile to a switch on the port"""

        if isinstance(mac_or_device, OmadaDevice):
            if mac_or_device.type != "switch":
                raise InvalidDevice()
            mac = mac_or_device.mac
        else:
            mac = mac_or_device

        if isinstance(index_or_port, OmadaSwitchPort):
            port = index_or_port
        else:
            port = await self.get_switch_port(mac_or_device, index_or_port)

        payload = {
            "name": new_name or port.name,
            "profileId": profile_id or port.profile_id,
            "profileOverrideEnable": not overrides is None,
        }
        if overrides:
            payload["operation"] = "switching"
            payload["bandWidthCtrlType"] = BandwidthControl.OFF
            payload["poe"] = (
                PoEMode.ENABLED if overrides.enable_poe else PoEMode.DISABLED
            )
            payload["dot1x"] = overrides.dot1x_mode
            payload["duplex"] = overrides.duplex
            payload["linkSpeed"] = overrides.link_speed
            payload["lldpMedEnable"] = overrides.lldp_med_enable
            payload["loopbackDetectEnable"] = overrides.loopback_detect
            payload["spanningTreeEnable"] = overrides.spanning_tree_enable
            payload["portIsolationEnable"] = overrides.port_isolation
            payload["topoNotifyEnable"] = False

        await self._api.request(
            "patch",
            self._api.format_url(f"switches/{mac}/ports/{port.port}", self._site_id),
            payload=payload,
        )

        # Read back the new port settings
        return await self.get_switch_port(mac, port)

    async def get_port_profiles(self) -> List[OmadaPortProfile]:
        """Lists the available switch port profiles that can be applied."""

        result = await self._api.request(
            "get", self._api.format_url("setting/lan/profileSummary", self._site_id)
        )

        return [OmadaPortProfile(p) for p in result["data"]]

    async def get_firmware_details(
        self, mac_or_device: Union[str, OmadaDevice]
    ) -> OmadaFirmwareUpdate:
        """Get details of the device firware and available upgrades."""

        if isinstance(mac_or_device, OmadaDevice):
            mac = mac_or_device.mac
        else:
            mac = mac_or_device

        result = await self._api.request(
            "get", self._api.format_url(f"devices/{mac}/firmware", self._site_id)
        )

        return OmadaFirmwareUpdate(result)

    async def start_firmware_upgrade(
        self, mac_or_device: Union[str, OmadaDevice]
    ) -> bool:
        """Begin an automatic firmware upgrade of the specified device"""

        if isinstance(mac_or_device, OmadaDevice):
            mac = mac_or_device.mac
        else:
            mac = mac_or_device

        payload = {"mac": mac}
        await self._api.request(
            "post",
            self._api.format_url(f"cmd/devices/{mac}/onlineUpgrade", self._site_id),
            payload=payload,
        )

        return True



""" Exceptions that the library might throw. """


class OmadaClientException(Exception):
    """Base for all exceptions raised by the library."""


class RequestFailed(OmadaClientException):
    """Generic rejection of any command by the controller."""

    def __init__(self, error_code: int, msg: str):
        self._error_code = error_code
        self._msg = msg
        super().__init__(f"Omada controller responded '{msg}' ({error_code})")


class LoginFailed(RequestFailed):
    """Username/Password failure."""


class LoginSessionClosed(OmadaClientException):
    """
    The login token isn't valid any more.

    If this happens immediately after logging on, and you are using IP addresses to contact the controller
    then make sure you supply a ClientSession that has an unsafe CookieJar, or the login session cookies
    won't work.
    """


class UnsupportedControllerVersion(OmadaClientException):
    """
    Indicates the Omada controller has a software version that is not supported.

    Only controller versions 5.0 and later are supported.
    """

    def __init__(self, version):
        super().__init__(f"Unsupported Omada controller version {version} found.")


class SiteNotFound(OmadaClientException):
    """The specified site cannot be found on the Controller."""


class ConnectionFailed(OmadaClientException):
    """Connection to Omada controller failed at the network level."""


class BadControllerUrl(OmadaClientException):
    """URL of controller could not be resolved."""


class InvalidDevice(OmadaClientException):
    """Device type isn't valid for this operation."""



"""Responsible for ~/.omada.cfg"""
from configparser import ConfigParser
from dataclasses import dataclass
from pathlib import Path

from tplink_omada_client.omadaclient import OmadaClient

_CONFIG_FILE: Path = Path.home() / '.omada.cfg'
_CONTROLLER_SECTION_PREFIX: str = 'controller:'
_CLI_SECTION: str = "cli"
_DEFAULT_TARGET: str = "default_target"

@dataclass
class ControllerConfig:
    """Holds config needed to issue calls to Omada controller"""
    url: str
    username: str
    password: str
    site: str

def get_targets() -> list[tuple[str, str, bool]]:
    """Get all the controllers in config file"""
    config_parser = _read_config_file()
    default_target = config_parser[_CLI_SECTION][_DEFAULT_TARGET] if _CLI_SECTION in config_parser else ""
    targets = []
    for (target, config) in config_parser.items():
        if target.startswith(_CONTROLLER_SECTION_PREFIX):
            name = target[len(_CONTROLLER_SECTION_PREFIX):]
            targets.append((name, config["url"], name == default_target))
    return targets

def get_target_config(name: str) -> ControllerConfig:
    """Using controller name to create Omada site client"""
    config = _read_config_file()

    if name == "":
        if _CLI_SECTION in config:
            name = config[_CLI_SECTION][_DEFAULT_TARGET]
        if not name:
            raise ValueError(f"No target specified, and no default target has been configured.")
        
    stored_name = _to_stored_name(name)
    if config.has_section(stored_name):
        stored_config = config[stored_name]
        return ControllerConfig(**stored_config)
    raise ValueError(f"Could not find target named '{name}'")

def set_target_config(name: str, config: ControllerConfig, set_default: bool) -> None:
    """Stores controller config in users's config file"""
    stored_name = _to_stored_name(name)
    config_parser = _read_config_file()
    config_parser.remove_section(stored_name)
    config_parser.add_section(stored_name)
    config_parser.set(stored_name, "url", config.url)
    config_parser.set(stored_name, "username", config.username)
    config_parser.set(stored_name, "password", config.password)
    config_parser.set(stored_name, "site", config.site)
    if set_default:
        if _CLI_SECTION not in config_parser:
            config_parser[_CLI_SECTION] = {}
        config_parser[_CLI_SECTION][_DEFAULT_TARGET] = name

    _write_config_file(config_parser)

def set_default_target(name: str) -> None:
    config_parser = _read_config_file()
    if _CLI_SECTION not in config_parser:
        config_parser[_CLI_SECTION] = {}
    config_parser[_CLI_SECTION][_DEFAULT_TARGET] = name
    _write_config_file(config_parser)

def to_omada_connection(config: ControllerConfig) -> OmadaClient:
    """Create a OmadaClient based on a ControllerConfig object"""
    return OmadaClient(config.url, username=config.username, password=config.password)

def _read_config_file() -> ConfigParser:
    """Read's the user's config file"""
    config = ConfigParser()
    if _CONFIG_FILE.exists():
        with _CONFIG_FILE.open() as file:
            config.read_file(file)
    return config

def _write_config_file(config_parser):
    """Write the user's config file"""
    with _CONFIG_FILE.open(mode='w') as file:
        config_parser.write(file)


def _to_stored_name(name: str) -> str:
    return _CONTROLLER_SECTION_PREFIX + name



"""
Definitions for Omada device objects

APs, Switches and Routers
"""
from typing import Any, List, Optional
from .definitions import (
    BandwidthControl,
    DeviceStatus,
    DeviceStatusCategory,
    Eth802Dot1X,
    LinkSpeed,
    LinkStatus,
    PoEMode,
    PortType,
)

class OmadaApiData:
    def __init__(self, data: dict[str, Any]):
        self._data = data

    @property
    def raw_data(self) -> dict[str, Any]:
        return self._data


class OmadaDevice(OmadaApiData):
    """Details of a device connected to the controller"""

    @property
    def type(self) -> str:
        """The type of the device. Its value can be "ap", "gateway", and "switch"."""
        return self._data["type"]

    @property
    def mac(self) -> str:
        """The MAC address of the device."""
        return self._data["mac"]

    @property
    def name(self) -> str:
        """The device name."""
        return self._data["name"]

    @property
    def model(self) -> str:
        """The device model, such as EAP225."""
        return self._data["model"]

    @property
    def model_display_name(self) -> str:
        """Model description for front-end display."""
        return self._data["showModel"]

    @property
    def status(self) -> DeviceStatus:
        """The status of the device."""
        return self._data["status"]

    @property
    def status_category(self) -> DeviceStatusCategory:
        """The high-level status of the device."""
        return self._data["statusCategory"]

    @property
    def ip_address(self) -> str:
        """IP address of the device."""
        return self._data["ip"]

    @property
    def display_uptime(self) -> str:
        """Uptime of the device, as a display string"""
        return self._data["uptime"]

    @property
    def uptime(self) -> int:
        """Uptime of the device, as a display string"""
        return self._data["uptimeLong"]

    @property
    def firmware_version(self) -> str:
        """Firmware version of the device"""
        return self._data["firmwareVersion"]


class OmadaListDevice(OmadaDevice):
    """An Omada Device (router, switch, eap) as represented in the device list"""

    @property
    def need_upgrade(self) -> bool:
        """True, if a firmware upgrade is available for the device."""
        return self._data["needUpgrade"]

    @property
    def fw_download(self) -> bool:
        """True, if a firmware upgrade is being downloaded."""
        return self._data["fwDownload"]


class OmadaLink(OmadaApiData):
    """Up/Downlink connection from a switch/ap device."""

    @property
    def mac(self) -> str:
        """The MAC of the linked device."""
        return self._data["mac"]

    @property
    def name(self) -> str:
        """The name of the linked device."""
        return self._data["name"]

    @property
    def type(self) -> str:
        """The type of device linked to."""
        return self._data["type"]

    @property
    def port(self) -> int:
        """The port's number."""
        return self._data["port"]


class OmadaDownlink(OmadaLink):
    """Downlink connection from a switch/ap port."""

    @property
    def type(self) -> str:
        """The type of device downlinked to."""
        return "ap"

    @property
    def model(self) -> str:
        """The model name of device linked to."""
        return self._data["model"]


class OmadaUplink(OmadaLink):
    """Uplink connection from a switch/ap device."""


class OmadaPortStatus(OmadaApiData):
    """Status information for a switch port."""

    @property
    def link_status(self) -> LinkStatus:
        """Port's link status."""
        return self._data["linkStatus"]

    @property
    def link_speed(self) -> LinkSpeed:
        """Port's link speed."""
        return self._data["linkSpeed"]

    @property
    def poe_active(self) -> bool:
        """Is the port powering a PoE device?"""
        return self._data["poe"]

    @property
    def poe_power(self) -> float:
        """Power (W) supplied over PoE."""
        if "poePower" in self._data:
            return self._data["poePower"]
        return 0.0

    @property
    def bytes_tx(self) -> int:
        """Number of bytes transmitted by the port."""
        return self._data["tx"]

    @property
    def bytes_rx(self) -> int:
        """Number of bytes received by the port."""
        return self._data["rx"]

    @property
    def stp_discarding(self) -> bool:
        """Stp blocking status in spanning tree."""
        return self._data["stpDiscarding"]


class OmadaSwitchPort(OmadaApiData):
    """Port on a switch/gateway device."""

    @property
    def port(self) -> int:
        """The port's number."""
        return self._data["port"]

    @property
    def name(self) -> str:
        """The device name."""
        return self._data["name"]

    @property
    def profile_id(self) -> str:
        """ID of the port's config profile."""
        return self._data["profileId"]

    @property
    def type(self) -> PortType:
        """The type of the port."""
        return self._data["type"]

    @property
    def operation(self) -> str:
        """Port config: switching, mirroring or aggregating."""
        return self._data["operation"]

    @property
    def is_disabled(self) -> bool:
        """Is the port disabled?"""
        return self._data["disable"]

    @property
    def port_status(self) -> OmadaPortStatus:
        """Status of the port."""
        return OmadaPortStatus(self._data["portStatus"])


class OmadaSwitchDeviceCaps(OmadaApiData):
    """Capabilities of a switch."""

    @property
    def poe_ports(self) -> int:
        """Number of PoE ports supported."""
        return self._data["poePortNum"]

    @property
    def supports_poe(self) -> bool:
        """Is PoE supported."""
        return self._data["poeSupport"]

    @property
    def supports_bt(self) -> bool:
        """Is BT supported."""
        return self._data["supportBt"]


class OmadaSwitch(OmadaDevice):
    """Details of a switch connected to the controller."""

    @property
    def number_of_ports(self) -> int:
        """The number of ports on the switch."""
        if "portNum" in self._data:
            return self._data["portNum"]
        # So much for the docs
        return self._data["deviceMisc"]["portNum"]

    @property
    def ports(self) -> List[OmadaSwitchPort]:
        """List of ports attached to the switch."""
        return [OmadaSwitchPort(p) for p in self._data["ports"]]

    @property
    def uplink(self) -> Optional[OmadaUplink]:
        """ Uplink device for this switch. """
        if not "uplink" in self._data:
            return None
        uplink = self._data["uplink"]
        if uplink is None:
            return None
        return OmadaUplink(uplink)

    @property
    def downlink(self) -> List[OmadaDownlink]:
        """Downlink devices attached to switch."""
        if "downlinkList" in self._data:
            return [OmadaDownlink(d) for d in self._data["downlinkList"]]
        return []

    @property
    def device_capabilities(self) -> OmadaSwitchDeviceCaps:
        """Capabilities of the switch."""
        return OmadaSwitchDeviceCaps(self._data["devCap"])


class OmadaAccesPointLanPortSettings(OmadaApiData):
    """A LAN port on an access point."""

    @property
    def port_name(self) -> str:
        """Name of the port - can't be edited"""
        return self._data["lanPort"]

    @property
    def supports_vlan(self) -> bool:
        """True if the port supports VLAN tagging"""
        return self._data["supportVlan"]

    @property
    def local_vlan_enable(self) -> bool:
        """True if VLAN tagging is enabled for the port explicitly"""
        return self._data["localVlanEnable"]

    @property
    def local_vlan_id(self) -> int:
        """VLAN ID for this port"""
        return self._data["localVlanId"]

    @property
    def supports_poe(self) -> bool:
        """True if the port supports PoE output"""
        return self._data["supportPoe"]

    @property
    def poe_enable(self) -> bool:
        """
        True to enable PoE.

        WARNING: Do not enable PoE for EAPs powered from another EAP
        """
        return self._data["poeOutEnable"]


class OmadaAccessPoint(OmadaDevice):
    """Details of an Access Point connected to the controller."""

    @property
    def wireless_linked(self) -> bool:
        """True, if the AP is connected wirelessley."""
        return self._data["wirelessLinked"]

    @property
    def supports_5g(self) -> bool:
        """True if 5G wifi is supported"""
        return self._data["deviceMisc"]["support5g"]

    @property
    def supports_5g2(self) -> bool:
        """True if 5G2 wifi is supported"""
        return self._data["deviceMisc"]["support5g2"]

    @property
    def supports_6g(self) -> bool:
        """True if Wifi 6 is supported"""
        return self._data["deviceMisc"]["support6g"]

    @property
    def supports_11ac(self) -> bool:
        """True if PoE is supported"""
        return self._data["deviceMisc"]["support11ac"]

    @property
    def supports_mesh(self) -> bool:
        """True if mesh networking is supported"""
        return self._data["deviceMisc"]["supportMesh"]

    @property
    def lan_port_settings(self) -> List[OmadaAccesPointLanPortSettings]:
        """Settings for the LAN ports on the access point"""
        return [
            OmadaAccesPointLanPortSettings(p) for p in self._data["lanPortSettings"]
        ]


class OmadaSwitchPortDetails(OmadaSwitchPort):
    """Full details of a port on a switch."""

    @property
    def port_id(self) -> str:
        """The ID of the port"""
        return self._data["id"]

    @property
    def max_speed(self) -> LinkSpeed:
        """The max speed of the port."""
        return self._data["maxSpeed"]

    @property
    def profile_name(self) -> str:
        """Name of the port's config profile"""
        return self._data["profileName"]

    @property
    def has_profile_override(self) -> bool:
        """True if the port's config profile has been overridden."""
        return self._data["profileOverrideEnable"]

    @property
    def poe_mode(self) -> PoEMode:
        """PoE config for this port."""
        return self._data["poe"]

    @property
    def bandwidth_limit_mode(self) -> BandwidthControl:
        """Type of bandwidth control applied."""
        return self._data["bandWidthCtrlType"]

    # "bandCtrl": {
    #     "egressEnable": false,
    #     "egressLimit": 0,
    #     "egressUnit": 1,
    #     "ingressEnable": false,
    #     "ingressLimit": 0,
    #     "ingressUnit": 1
    # },
    # "stormCtrl": {
    #     "unknownUnicastEnable": false,
    #     "unknownUnicast": 0,
    #     "multicastEnable": false,
    #     "multicast": 0,
    #     "broadcastEnable": false,
    #     "broadcast": 0,
    #     "action": 0,
    #     "recoverTime": 3600
    # },

    @property
    def eth_802_1x_control(self) -> Eth802Dot1X:
        """802.1x Auth mode"""
        return self._data["dot1x"]

    @property
    def lldp_med_enabled(self) -> bool:
        """LLDP Mode"""
        return self._data["lldpMedEnable"]

    @property
    def topology_notify_enabled(self) -> bool:
        """Topology notify mode"""
        return self._data["topoNotifyEnable"]

    @property
    def spanning_tree_enabled(self) -> bool:
        """Spanning tree loopback control"""
        return self._data["spanningTreeEnable"]

    @property
    def loopback_detect_enabled(self) -> bool:
        """Loopback detection"""
        return self._data["loopbackDetectEnable"]

    @property
    def port_isolation_enabled(self) -> bool:
        """Port isolation (Danger!)"""
        return self._data["portIsolationEnable"]


class OmadaPortProfile(OmadaApiData):
    """Definition of a switch port configuration profile."""

    @property
    def profile_id(self) -> str:
        """ID of this profile."""
        return self._data["id"]

    @property
    def site(self) -> str:
        """Site which this profile is valid for."""
        return self._data["site"]

    @property
    def name(self) -> str:
        """Name of the profile."""
        return self._data["name"]

    @property
    def poe_mode(self) -> PoEMode:
        """PoE mode."""
        return self._data["poe"]

    @property
    def bandwidth_limit_mode(self) -> BandwidthControl:
        """Type of bandwidth control applied."""
        return self._data["bandWidthCtrlType"]

    @property
    def eth_802_1x_control(self) -> Eth802Dot1X:
        """802.1x Auth mode"""
        return self._data["dot1x"]

    @property
    def lldp_med_enabled(self) -> bool:
        """LLDP Mode"""
        return self._data["lldpMedEnable"]

    @property
    def topology_notify_enabled(self) -> bool:
        """Topology notify mode"""
        return self._data["topoNotifyEnable"]

    @property
    def spanning_tree_enabled(self) -> bool:
        """Spanning tree loopback control"""
        return self._data["spanningTreeEnable"]

    @property
    def loopback_detect_enabled(self) -> bool:
        """Loopback detection"""
        return self._data["loopbackDetectEnable"]

    @property
    def port_isolation_enabled(self) -> bool:
        """Port isolation (Danger!)"""
        return self._data["portIsolationEnable"]


class OmadaInterfaceDetails(OmadaApiData):
    """Basic UI Information about controller."""

    @property
    def controller_name(self) -> str:
        """Display name of the controller."""
        return self._data["controllerName"]


class OmadaFirmwareUpdate(OmadaApiData):
    """Status information for a switch port."""

    @property
    def current_version(self) -> str:
        """Device's current firmware version."""
        return self._data["curFwVer"]

    @property
    def latest_version(self) -> str:
        """Latest firmware version available."""
        return self._data["lastFwVer"]

    @property
    def release_notes(self) -> str:
        """Release notes for the new firmware."""
        return self._data["fwReleaseLog"]



"""Internal Omada API client."""

import time
from typing import Any, Optional, Tuple

import re
from urllib.parse import urlsplit, urljoin
from aiohttp import client_exceptions, CookieJar
from aiohttp.client import ClientSession
from awesomeversion  import AwesomeVersion

from .exceptions import (
    BadControllerUrl,
    ConnectionFailed,
    LoginFailed,
    LoginSessionClosed,
    RequestFailed,
    UnsupportedControllerVersion,
)


class OmadaApiConnection:
    """Low level Omada API client."""

    _own_session: bool
    _controller_id: str
    _controller_version: str
    _csrf_token: Optional[str]
    _last_logon: float

    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        websession: Optional[ClientSession] = None,
        verify_ssl=True,
    ):

        if not url.lower().startswith(("http://", "https://")):
            url = "https://" + url
        url_parts = urlsplit(url, "https://")
        self._url = url_parts.geturl()
        self._host = url_parts.hostname or ""

        self._url = url
        self._username = username
        self._password = password
        self._session = websession
        self._verify_ssl = verify_ssl
        self._csrf_token = None

    async def _get_session(self) -> ClientSession:
        if self._session is None:
            self._own_session = True
            jar = (
                None
                if re.fullmatch(r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", self._host)
                is None
                else CookieJar(unsafe=True)
            )
            self._session = ClientSession(cookie_jar=jar)
        return self._session

    async def __aenter__(self):
        try:
            await self.login()
            return self
        except Exception as error:
            if self._own_session:
                await self.close()
            raise error

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> bool:
        """Call when the client is disposed."""
        # Close the web session, if we created it (i.e. it was not passed in)
        if self._own_session:
            await self.close()
        return False

    async def close(self):
        """Close the current web session."""
        if self._session:
            await self._session.close()
            self._session = None

    async def login(self) -> str:
        """
        Call to obtain login token, controller id, and site id.

        Calls to login are optional, as the API will automatically authenticate as necessary.
        However, you may want to attempt a login to check connectivity.
        """

        version, controller_id = await self._get_controller_info()

        if AwesomeVersion(version) < AwesomeVersion("5.1.0"):
            raise UnsupportedControllerVersion(self._controller_version)

        self._controller_id = controller_id
        self._controller_version = version

        auth = {"username": self._username, "password": self._password}
        response = await self._do_request(
            "post", self.format_url("login"), payload=auth
        )

        self._csrf_token = response["token"]
        self._last_logon = time.time()

        return self._controller_id

    async def _check_login(self) -> bool:
        if not self._csrf_token:
            return False

        if time.time() - self._last_logon < 60 * 60:
            # Assume 1hr is good for a login to remain active
            return True

        try:
            response = await self._do_request("get", self.format_url("loginStatus"))
            logged_in = bool(response["login"])
            if logged_in:
                self._last_logon = time.time()
            return logged_in
        except:  # pylint: disable=bare-except
            return False

    async def _get_controller_info(self) -> Tuple[str, str]:
        """Get Omada controller version and Id (unauthenticated)."""

        response = await self._do_request("get", urljoin(self._url, "/api/info"))

        return (response["controllerVer"], response["omadacId"])

    def format_url(self, end_point: str, site: Optional[str] = None) -> str:
        """Get a REST url for the controller action"""

        if site:
            end_point = f"sites/{site}/{end_point}"

        return urljoin(self._url, f"/{self._controller_id}/api/v2/{end_point}")

    async def request(self, method: str, url: str, params=None, payload=None) -> Any:
        """Perform a request specific to the controlller, with authentication"""

        if not await self._check_login():
            await self.login()

        return await self._do_request(method, url, params=params, payload=payload)

    async def _do_request(
        self, method: str, url: str, params=None, payload=None
    ) -> Any:
        """Perform a request on the controller, and unpack the response."""

        session = await self._get_session()

        # Note: Auth happens via cookies, set during the login command, but we also get a CSRF token
        # which we need to push back
        headers = {}
        if self._csrf_token:
            headers["Csrf-Token"] = self._csrf_token

        try:
            async with session.request(
                method,
                url,
                params=params,
                headers=headers,
                json=payload,
                ssl=self._verify_ssl,
            ) as response:

                if response.status != 200:
                    if response.content_type == "application/json":
                        content = await response.json(encoding="utf-8")
                        self._check_application_errors(content)

                    raise RequestFailed(response.status, "HTTP Request Error")

                # If something goes wrong with the login session, Omada requests return "success", and a login page. :/
                if response.content_type != "application/json":
                    raise LoginSessionClosed()

                content = await response.json(encoding="utf-8")
                self._check_application_errors(content)

                # Unpack response data
                if "result" in content:
                    return content["result"]
                return content

        except client_exceptions.InvalidURL as err:
            raise BadControllerUrl(err) from err
        except client_exceptions.ClientConnectionError as err:
            raise ConnectionFailed(err) from err
        except client_exceptions.ClientError as err:
            raise RequestFailed(0, f"Unexpected error: {err}") from None

    def _check_application_errors(self, response):
        if not isinstance(response, dict):
            return
        if response["errorCode"] == 0:
            return
        if response["errorCode"] == -30109:
            raise LoginFailed(response["errorCode"], response["msg"])
        raise RequestFailed(response["errorCode"], response["msg"])



"""Implementation of 'default' command"""
from argparse import _SubParsersAction, ArgumentError
import getpass

from .util import assert_target_argument
from .config import ControllerConfig, set_default_target, to_omada_connection, get_target_config

async def command_target(args) -> int:
    """Executes 'default' command"""
    target = assert_target_argument(args)
    config = get_target_config(target)

    set_default_target(target)
    return 0

def arg_parser(subparsers: _SubParsersAction) -> None:
    """Configures argument parser for 'default' command"""
    set_parser = subparsers.add_parser(
        "default",
        help="Sets the default target",
    )
    set_parser.set_defaults(func=command_target)



""" Simple Http client for Omada controller REST api. """
from typing import NamedTuple, Optional, Union
from aiohttp.client import ClientSession

from .omadasiteclient import OmadaSiteClient
from .omadaapiconnection import OmadaApiConnection


from .exceptions import (
    SiteNotFound,
)
from .devices import (
    OmadaInterfaceDetails,
)


class OmadaSite(NamedTuple):
    """Identifies a site controlled by the controller."""

    name: str
    id: str


class OmadaClient:
    """
    Simple client for Omada controller API

    Provides a very limited subset of the API documented in the
    'Omada_SDN_Controller_V5.0.15 API Document'
    """

    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        websession: Optional[ClientSession] = None,
        verify_ssl=True,
    ):
        self._api = OmadaApiConnection(url, username, password, websession, verify_ssl)

    async def __aenter__(self):
        await self._api.__aenter__()
        return self

    async def __aexit__(self, *args) -> bool:
        """Call when the client is disposed."""
        # Close the web session, if we created it (i.e. it was not passed in)
        return await self._api.__aexit__(*args)

    async def login(self) -> str:
        """
        Log in to the controller and returns the controller's unique ID.

        Calls to login are optional, as the API will automatically authenticate as necessary.
        However, you may want to attempt a login to check connectivity.
        """
        return await self._api.login()

    async def get_controller_name(self) -> str:
        """Get the display name of the Omada controller."""
        result = await self._api.request(
            "get", self._api.format_url("maintenance/uiInterface")
        )

        return OmadaInterfaceDetails(result).controller_name

    async def get_sites(self) -> list[OmadaSite]:
        """Get basic list of sites the user can see"""
        response = await self._api.request("get", self._api.format_url("users/current"))

        sites = [OmadaSite(s["name"], s["key"]) for s in response["privilege"]["sites"]]
        return sites

    async def get_site_client(self, site: Union[str, OmadaSite]) -> OmadaSiteClient:
        """Get a client that can query the specified Omada site."""
        if isinstance(site, OmadaSite):
            site_id = site.id
        else:
            site_id = await self._get_site_id(site)
        return OmadaSiteClient(site_id, self._api)

    async def _get_site_id(self, site_name: str):
        """Get site id by (display) name"""

        # The current user object has a list of allowed sites to administer
        sites = await self.get_sites()

        site_id = next((s.id for s in sites if s.name == site_name), None)

        if site_id:
            return site_id

        raise SiteNotFound(f"Site '{site_name}' not found")



"""Implementation of 'targets' command"""
from argparse import _SubParsersAction

from .config import (
    get_targets,
)

async def command_targets(args) -> int: # pylint: disable=unused-argument
    """Executes 'targets' command"""
    controllers = get_targets()
    for (controller, url, is_default) in controllers:
        print(f"{('*' if is_default else' ')} {controller:15} {url}")
    return 0

def arg_parser(subparsers: _SubParsersAction) -> None:
    """Configures argument parser for 'targets' command"""
    # targets
    controllers_parser = subparsers.add_parser(
        "targets",
        aliases=['t'],
        help="Lists the configured targets")
    controllers_parser.set_defaults(func=command_targets)



"""Implementation for 'devices' command"""

from argparse import _SubParsersAction
import json
from .config import get_target_config, to_omada_connection
from .util import dump_raw_data, get_target_argument

async def command_devices(args) -> int:
    """Executes 'devices' command"""
    controller = get_target_argument(args)
    config = get_target_config(controller)

    async with to_omada_connection(config) as client:
        site_client = await client.get_site_client(config.site)
        for device in await site_client.get_devices():
            print(f"{device.mac} {device.ip_address:>15} {device.type:>7}  {device.name:20} {device.model_display_name}")
            dump_raw_data(args, device)
    return 0

def arg_parser(subparsers: _SubParsersAction) -> None:
    """Configures arguments parser for 'devices' command"""
    devices_parser = subparsers.add_parser(
        "devices",
        aliases=['d'],
        help="Lists devices managed by Omada Controller")
    devices_parser.add_argument('-d', '--dump', help="Output raw device information",  action='store_true')

    devices_parser.set_defaults(func=command_devices)



""" Omada CLI """
import argparse
import asyncio
import sys
from typing import Any, Sequence, Union

from tplink_omada_client.exceptions import LoginFailed

from . import (
    command_devices,
    command_switch,
    command_switches,
    command_target,
    command_targets,
    command_default
)

def main(argv: Union[Sequence[str], None] = None) -> int:
    """Entry point for Omada CLI"""
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-t', '--target',
        help="The target Omada controller",
        default=""
    )

    subparsers = parser.add_subparsers(
        title='Available commands',
        metavar='command',
    )

    command_devices.arg_parser(subparsers)
    command_switch.arg_parser(subparsers)
    command_switches.arg_parser(subparsers)
    command_target.arg_parser(subparsers)
    command_targets.arg_parser(subparsers)
    command_default.arg_parser(subparsers)

    try:
        args = parser.parse_args(args=argv)
        if "func" in args:
            return asyncio.run(args.func(vars(args)))
        parser.print_help()
        return 1
    except argparse.ArgumentError as error:
        parser.print_usage()
        print(error.message)
        return 1
    except LoginFailed as error:
        print(error)
        return 2



"""Implementation for 'switches' command"""

from argparse import _SubParsersAction

from .config import get_target_config, to_omada_connection
from .util import dump_raw_data, get_target_argument

async def command_switches(args) -> int:
    """Executes 'switches' command"""
    controller = get_target_argument(args)
    config = get_target_config(controller)

    async with to_omada_connection(config) as client:
        site_client = await client.get_site_client(config.site)
        for switch in await site_client.get_switches():
            print(f"{switch.mac} {switch.ip_address:>15}  {switch.name:20} ", end="")
            for port in switch.ports:
                if port.is_disabled:
                    print("x", end="")
                elif port.port_status.link_status:
                    print("\u2611", end="")
                else:
                    print("\u2610", end="")
                if port.port_status.poe_active:
                    print("\u26a1", end="")
                else:
                    print("  ", end="")

            dump_raw_data(args, switch)
            print()
    return 0

def arg_parser(subparsers: _SubParsersAction) -> None:
    """Configures arguments parser for 'switches' command"""
    switches_parser = subparsers.add_parser(
        "switches",
        aliases=['s'],
        help="Lists switches managed by Omada Controller")
    switches_parser.set_defaults(func=command_switches)
    switches_parser.add_argument('-d', '--dump', help="Output raw device information",  action='store_true')



"""Common functionality for multiple commands"""
import argparse
import json

from typing import Any
from re import IGNORECASE, match
from tplink_omada_client.devices import OmadaApiData
from tplink_omada_client.omadasiteclient import OmadaSiteClient

TARGET_ARG: str = "target"

def assert_target_argument(args: dict[str, Any]) -> str:
    """Throws ArgumentError if target arg missing"""
    if args[TARGET_ARG] == "": # The default is now empty
        raise argparse.ArgumentError(None, f"error: Missing --{TARGET_ARG} argument")
    return args[TARGET_ARG]

def get_target_argument(args: dict[str, Any]) -> str:
    return args[TARGET_ARG]

async def get_mac(site_client: OmadaSiteClient, mac_or_name: str) -> str:
    if match('([0-9A-F]{2}[-]){5}[0-9A-F]{2}$',
        string=mac_or_name,
        flags=IGNORECASE):
        return mac_or_name

    device = next((d for d in await site_client.get_devices() if d.name == mac_or_name), None)
    if not device:
        raise argparse.ArgumentError(None, f"Device with name {mac_or_name} not found")
    return device.mac

def dump_raw_data(args: dict[str, Any], data: OmadaApiData):
    if args['dump']:
        print("--- BEGIN RAW DATA ---")
        print(json.dumps(data.raw_data, indent=2,ensure_ascii=False))
        print("---  END RAW DATA  ---")



#!/usr/bin/env python3
""" Cmd line to work out the API. """

import asyncio
import sys
from pprint import pprint
from src.tplink_omada_client.omadaclient import OmadaClient
from src.tplink_omada_client.omadasiteclient import AccessPointPortSettings


async def do_the_magic(url: str, username: str, password: str):
    """Not a real test client."""

    async with OmadaClient(url, username, password, verify_ssl=False) as client:

        print(f"Found Omada Controller: {await client.get_controller_name()}")

        sites = await client.get_sites()
        print(f"Found {len(sites)} sites")
        for site in sites:
            print(f"Connecting to {site.name}")

            site_client = await client.get_site_client(sites[0])

            devices = await site_client.get_devices()

            print(f"Found {len(devices)} Omada devices in site {site.name}.")
            print(f"    {len([d for d in devices if d.type == 'ap'])} Access Points.")
            print(f"    {len([d for d in devices if d.type == 'switch'])} Switches.")
            print(f"    {len([d for d in devices if d.type == 'gateway'])} Routers.")

            for firmware_details in [
                await site_client.get_firmware_details(d)
                for d in devices
                if d.need_upgrade
            ]:
                print("Available firmware upgrade:")
                pprint(vars(firmware_details))

            access_points = [
                await site_client.get_access_point(a) for a in devices if a.type == "ap"
            ]
            for access_point in access_points:
                print(f"Access Point: {access_point.name}")
                if access_point.name == "Office":
                    port_status = await site_client.update_access_point_port(
                        access_point, "ETH3", AccessPointPortSettings(enable_poe=True)
                    )
                    pprint(vars(port_status))
                    port_status = await site_client.update_access_point_port(
                        access_point, "ETH3", AccessPointPortSettings(enable_poe=False)
                    )
                    pprint(vars(port_status))

            # pprint(vars(devices[0]))

            # Get full info of all switches
            switches = [
                await site_client.get_switch(s) for s in devices if s.type == "switch"
            ]

            # pprint(vars(switches[0]))

            # ports = await client.get_switch_ports(switches[0])

            port = await site_client.get_switch_port(switches[0], switches[0].ports[4])
            print(f"Port index 4: {port.name} Profile: {port.profile_name}")
            pprint(vars(port))

            updated_port = await site_client.update_switch_port(
                switches[0], port, new_name="Port5"
            )
            pprint(vars(updated_port))

            profiles = await site_client.get_port_profiles()
            pprint(vars(profiles[0]))

        print("Done.")


def main():
    """Basic sample test client."""
    if len(sys.argv) < 2 or len(sys.argv) > 4:
        print("Usage: client <omada url> [username] [password]")
        return

    omadaurl = sys.argv[1]
    username = "admin" if len(sys.argv) < 3 else sys.argv[2]
    password = "admin" if len(sys.argv) < 4 else sys.argv[3]

    asyncio.run(do_the_magic(omadaurl, username, password))


main()



"""Implementation of 'target' command"""
from argparse import _SubParsersAction, ArgumentError
import getpass
import json

from .util import get_target_argument
from .config import ControllerConfig, set_target_config, to_omada_connection

async def command_target(args) -> int:
    """Executes 'target' command"""
    target = get_target_argument(args)

    if args['password']:
        password = args['password']
    else:
        password = getpass.getpass()
    config = ControllerConfig(
        url=args['url'],
        username=args['username'],
        password=password,
        site=args['site'],
    )
    
    # Connect to controller to validate config
    async with to_omada_connection(config) as client:
        name = await client.get_controller_name()
        for site in await client.get_sites():
            if args['site'] == site.name:
                print(f"Set target {target} to controller {name} and site {site.name}")
                set_target_config(target, config, args['set_default'])

    return 0

def arg_parser(subparsers: _SubParsersAction) -> None:
    """Configures argument parser for 'target' command"""
    set_parser = subparsers.add_parser(
        "target",
        help="Add Omada Controller to list of targets",
    )
    set_parser.set_defaults(func=command_target)
    set_parser.add_argument(
        '--url',
        help="The URL of the Omada controller",
        required=True,
    )
    set_parser.add_argument(
        '--username',
        help="The name of the user used to authenticate",
        required=True,
    )
    set_parser.add_argument(
        '--password',
        help="The user's password, password will be prompted if not provided",
    )
    set_parser.add_argument(
        '--site',
        help="The Omada site to user",
        default = "Default",
    )
    set_parser.add_argument(
        '-sd',
        '--set-default',
        help="Set this target as the default",
        action='store_true')




"""Implementation for 'switch' command"""

from argparse import ArgumentParser

from .config import get_target_config, to_omada_connection
from .util import dump_raw_data, get_mac, get_target_argument

async def command_switch(args) -> int:
    """Executes 'switch' command"""
    controller = get_target_argument(args)
    config = get_target_config(controller)

    async with to_omada_connection(config) as client:
        site_client = await client.get_site_client(config.site)
        mac = await get_mac(site_client, args['mac'])
        switch = await site_client.get_switch(mac)
        print(f"Name: {switch.name}")
        print(f"Address: {switch.mac} ({switch.ip_address})")
        print(f"Ports: {switch.number_of_ports}")
        print(f"Supports PoE: {switch.device_capabilities.supports_poe}")
        if switch.device_capabilities.supports_poe:
            print(f"PoE ports: {switch.device_capabilities.poe_ports}")
        print(f"Model: {switch.model_display_name}")
        switch.status_category
        print(f"Uptime: {switch.display_uptime}")
        if switch.uplink:
            print(f"Uplink switch: {switch.uplink.mac} {switch.uplink.name}")
        if len(switch.downlink) > 0:
            print("Downlink devices:")
            for downlink in switch.downlink:
                print(f"- {downlink.mac} {downlink.name}")

        dump_raw_data(args, switch)

    return 0

def arg_parser(subparsers) -> None:
    """Configures arguments parser for 'switches' command"""
    switch_parser: ArgumentParser = subparsers.add_parser(
        "switch",
        help="Shows details about the specified switch"
    )
    switch_parser.set_defaults(func=command_switch)

    switch_parser.add_argument(
        "mac",
        help="The MAC address or name of the switch",
    )
    switch_parser.add_argument('-d', '--dump', help="Output raw device information",  action='store_true')

